{
  open Parser
  open Printf
  exception Lexing_error of string
  exception Unexpected_eof

  let doc_buffer = Buffer.create 100
  let string_buffer = Buffer.create 100
  let fead_buffer b c = Buffer.add_char b c
  let read_buffer b =
    let str = Buffer.contents b in
      Buffer.clear b;
      str
}

let newline   = ( '\010' | '\013' | "\013\010" )
let space     = ( ' ' | '\t' )
let letter    = [ 'a' - 'z' 'A' - 'Z' ]
let word      = ( letter | '_' ) ( letter | '_' | [ '0' - '9' ] )*
let digit     = [ '0' - '9' ]
let sign      = [ '-' '+' ]?
let intbin    = "0b" [ '0' '1' ]+
let intdec    = digit+
let inthex    = "0x" [ '0' - '9' 'a' - 'f' 'A' - 'F' ]+
let integer   = intbin | intdec | inthex
let floatting = digit+ '.' digit+ ('e' sign digit+)?

rule main = parse
  | newline   { Lexing.new_line lexbuf ; main lexbuf }
  | space     { main lexbuf }

  | '"'       { in_string lexbuf      }
  | "/**"     { in_doc_comment lexbuf }
  | "/*"      { in_c_comment lexbuf   }
  | "//"      { in_cpp_comment lexbuf }

  | '.'       { DOT       }
  | ','       { COMMA     }
  | ':'       { COLON     }
  | ';'       { SEMICOLON }

  | '='       { EQUAL         }
  | '+'       { PLUS          }
  | '-'       { MINUS         }
  | '*'       { MUL           }
  | '/'       { DIV           }
  | '%'       { MOD           }
  | '|'       { OR            }
  | '&'       { AND           }
  | '^'       { XOR           }
  | '~'       { TILD          }
  | "<<"      { INF_INF       }
  | ">>"      { SUP_SUP       }
  | ">>>"     { SUP_SUP_SUP   }
  | "&&"      { AND_AND       }
  | "||"      { OR_OR         }
  | "+="      { PLUS_EQUAL    }
  | "-="      { MINUS_EQUAL   }
  | "*="      { MUL_EQUAL     }
  | "/="      { DIV_EQUAL     }
  | "%="      { MOD_EQUAL     }
  | "|="      { OR_EQUAL      }
  | "&="      { AND_EQUAL     }
  | "^="      { XOR_EQUAL     }
  | "~="      { TILD_EQUAL    }
  | "<<="     { INF_INF_EQUAL }
  | ">>="     { SUP_SUP_EQUAL }
  | ">>>="    { SUP_SUP_SUP_EQUAL }
  | "&&="     { AND_AND_EQUAL }
  | "||="     { OR_OR_EQUAL   }
  | "++"      { PLUS_PLUS     }
  | "--"      { MINUS_MINUS   }
  | "=="      { EQUAL_EQUAL   }
  | "!="      { NOT_EQUAL     }
  | '>'       { SUP           }
  | ">="      { SUP_EQUAL     }
  | '<'       { INF           }
  | "<="      { INF_EQUAL     }
  | '!'       { NOT           }
  | '?'       { INTERG        }

  | '('       { LEFTPAR      }
  | ')'       { RIGHTPAR     }
  | '['       { LEFTBRACKET  }
  | ']'       { RIGHTBRACKET }
  | '{'       { LEFTBRACE    }
  | '}'       { RIGHTBRACE   }

  | word as w { match w with
		  | "package"      -> PACKAGE
		  | "import"       -> IMPORT
		  | "class"        -> CLASS
		  | "extends"      -> EXTENDS
		  | "implements"   -> IMPLEMENTS
		  | "interface"    -> INTERFACE
		  | "return"       -> RETURN
		  | "throw"        -> THROW
		  | "break"        -> BREAK
		  | "continue"     -> CONTINUE
		  | "if"           -> IF
		  | "else"         -> ELSE
		  | "do"           -> DO
		  | "while"        -> WHILE
		  | "for"          -> FOR
		  | "try"          -> TRY
		  | "catch"        -> CATCH
		  | "finally"      -> FINALLY
		  | "switch"       -> SWITCH
		  | "case"         -> CASE
		  | "default"      -> DEFAULT
		  | "null"         -> NULL
		  | "super"        -> SUPER
		  | "this"         -> THIS
		  | "instanceof"   -> INSTANCEOF
		  | "true"         -> TRUE
		  | "false"        -> FALSE
		  | "new"          -> NEW
		  | "boolean"      -> BOOLEAN
		  | "byte"         -> BYTE
		  | "char"         -> CHAR
		  | "short"        -> SHORT
		  | "int"          -> INT
		  | "float"        -> FLOAT
		  | "long"         -> LONG
		  | "double"       -> DOUBLE
		  | "public"       -> PUBLIC
		  | "private"      -> PRIVATE
		  | "protected"    -> PROTECTED
		  | "static"       -> STATIC
		  | "final"        -> FINAL
		  | "native"       -> NATIVE
		  | "synchronized" -> SYNCHRONIZED
		  | "abstract"     -> ABSTRACT
		  | "threadsafe"   -> THREADSAFE
		  | "transient"    -> TRANSIENT
		  | "volatile"     -> VOLATILE
		  | "throws"       -> THROWS
		  | "void"         -> VOID
		  | _ -> IDENT (w)
	      }
  | "System.out.println" { PRINTLN }
  | integer as i {
      try INTEGER (int_of_string i) with _ ->
	let msg = sprintf "invalid integer: %s" i in
	  raise (Lexing_error msg)
    }
  | floatting as f {
      try FLOATING (float_of_string f) with _ ->
	let msg = sprintf "invalid float: %s" f in
	  raise (Lexing_error msg)
    }
  | '\'' (_ as c) '\'' { CHARACTER c    }
  | "'\\'"             { CHARACTER '\'' }
  | "'\\n'"            { CHARACTER '\n' }
  | "'\\t'"            { CHARACTER '\t' }
  | "'\\r'"            { CHARACTER '\r' }
  | "'\\\\'"           { CHARACTER '\\' }
  | eof    { EOF }
  | _ as c {
      let msg = sprintf "invalid character: '%c'" c in
	raise (Lexing_error msg)
    }

and in_string = parse
  | '"'          { STRING (read_buffer string_buffer) }
  | '\\' newline { Lexing.new_line lexbuf ; in_string lexbuf }
  | newline { Lexing.new_line lexbuf;
	      fead_buffer string_buffer '\n';
	      in_string lexbuf }
  | "\\\""  { fead_buffer string_buffer '"' ; in_string lexbuf }
  | "\\n"   { fead_buffer string_buffer '\n' ; in_string lexbuf }
  | "\\t"   { fead_buffer string_buffer '\t' ; in_string lexbuf }
  | "\\r"   { fead_buffer string_buffer '\r' ; in_string lexbuf }
  | "\\\\"  { fead_buffer string_buffer '\\' ; in_string lexbuf }
  | _ as c  { fead_buffer string_buffer c ; in_string lexbuf }
  | eof     { raise Unexpected_eof }

and in_doc_comment = parse
  | "*/"    { DOC (read_buffer doc_buffer) }
  | newline { Lexing.new_line lexbuf;
	      fead_buffer doc_buffer '\n';
	      in_doc_comment lexbuf
	    }
  | _ as c  { fead_buffer doc_buffer c ; in_doc_comment lexbuf }
  | eof     { raise Unexpected_eof }

and in_c_comment = parse
  | "*/"    { main lexbuf }
  | newline { Lexing.new_line lexbuf ; in_c_comment lexbuf }
  | _       { in_c_comment lexbuf }
  | eof     { raise Unexpected_eof }

and in_cpp_comment = parse
  | newline { Lexing.new_line lexbuf ; main lexbuf }
  | _       { in_cpp_comment lexbuf }
  | eof     { EOF }
