let print_error lexbuf msg =
  let start = lexbuf.Lexing.lex_start_p in
  let curr = lexbuf.Lexing.lex_curr_p in
  let first_char = start.Lexing.pos_cnum - start.Lexing.pos_bol in
  let last_char = curr.Lexing.pos_cnum - start.Lexing.pos_bol in
    if first_char + 1 < last_char then
      Printf.eprintf "Error: line %d, characters %d-%d: %s.\n%!"
        start.Lexing.pos_lnum first_char last_char msg
    else
      Printf.eprintf "Error: line %d, character %d: %s.\n%!"
        start.Lexing.pos_lnum first_char msg;
;;

let output = ref "a.out" in
let overmeth = ref "java_1_5" in
let verbose = ref true in
let printast = ref false in
let printtast = ref false in
let simulate = ref false in
let compile = ref true in
let src = ref None in
let spec =
  Arg.align [
    ("-o", Arg.Set_string output,
     "<filename> Define output file (default: a.out)");
    ("-verbose", Arg.Set verbose, " Verbose mode (default)");
    ("-silent", Arg.Clear verbose, " Silent mode");
    ("-print-ast", Arg.Set printast, " Print the Abstract Syntax Tree");
    ("-print-typed-ast", Arg.Set printtast,
     " Print the typed Abstract Syntax Tree");
    ("-simulate", Arg.Set simulate, " Interpret code in a simulator");
    ("-no-compile", Arg.Clear compile, " Do not compile code");
    ("-overloader", Arg.Set_string overmeth,
     "<technique>  Define overloading discipline");
  ]
in
let usage = Printf.sprintf "Usage: %s [ OPTIONS ] <Src.java>" Sys.argv.(0) in
let error msg =
  Printf.printf "Error: %s\n" msg;
  Arg.usage spec usage;
  exit 1;
in
let unknow arg =
  match !src with
    | None -> src := Some arg
    | Some _ -> error (Printf.sprintf "don't know what to do with `%s'" arg)
in
  Arg.parse spec unknow usage;
  let src =
    match !src with
      | None -> error "source file undefined"
      | Some x -> x
  in
  let overloader =
    match !overmeth with
      | "first"    -> Overloader.First
      | "java_1_2" -> Overloader.Java_1_2
      | "java_1_5" -> Overloader.Java_1_5
      | "mult"     -> Overloader.Multiply
      | "perfect"  -> Overloader.Perfect
      | x -> error (Printf.sprintf  "unknow overloading method: %s" x)
  in
  let lexbuf =
    try
      Lexing.from_channel (open_in src)
    with Sys_error _ -> error (Printf.sprintf "failed to open file `%s'" src)
  in
    try
      let cu = Parser.main Lexer.main lexbuf in
      if !printast then Printer.print_compil_unit stdout cu;
      let tcu = Typer.type_compil_unit cu !verbose overloader in
      if !printtast then Printer.print_typed_compil_unit stdout tcu;
      let _ = overloader in
	if !simulate then Interpreter.eval_compil_unit tcu;
	if !compile then (
	  Codegen.initialise_code ();
	  Codegen.gen_compil_unit tcu;
	  Codegen.finalise_code ();
	  Linker.export !output;
	)
    with
      | Lexer.Lexing_error msg ->
	  print_error lexbuf msg;
	  exit 1;
      | Parsing.Parse_error ->
	  print_error lexbuf "syntax error";
	  exit 1;
      | Lexer.Unexpected_eof ->
	  print_error lexbuf "unexpected end of file";
	  exit 1;
      | Ast_typer.Type_error(_) as te -> Printer.print_type_error stderr te;
;;
